//------------------------------ $Keywords ----------------------------------
// SelfImage - Disk image manipulation program
// SelfImage_TImageStore.cpp - TImageStore class
// Copyright 2005, Kurt Fitzner <kfitzner@excelcia.org>
//---------------------------------------------------------------------------
// This file is part of SelfImage.
//
// SelfImage is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License (Version 2) as
// published by the Free Software Foundation.
//
// SelfImage is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// VCS: $Version: 0 $ $Revision: 3 $
/*
$History: **** V 0.1 by kfitzner ****
$History: * selfimage_timagestore.cpp - 2005-11-07 2:39:39 AM - 8573 Bytes
$History: * selfimage_timagestore.h - 2005-10-21 10:56:46 AM - 2239 Bytes
$History: * Initial check-in
$History: **** V 0.2 by kfitzner ****
$History: * selfimage_timagestore.cpp - 2005-11-12 5:22:11 PM - 8837 Bytes
$History: * selfimage_timagestore.h - 2005-11-12 5:19:26 PM - 2243 Bytes
$History: * Typo in copyright area + change in program description
$History: **** Latest ** V 0.3 by kfitzner ** 2005-11-27 6:09:23 AM ****
$History: * Add decompression support
*/
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// File Notes:
//---------------------------------------------------------------------------
// 12 Sep 2005 - Kurt Fitzner <kfitzner@excelcia.org>
//
// An image "store" represtents any place where, no surprise, an image can
// be stored or retrieved from.  A file, a partition, even a network source.
// This class abstracts opens, closes, reads, etc so that the other classes
// don't need to know (or care) where their data is coming from or going.
//
// This class doesn't need to be particularly thread-safe, since it's a
// little unlikely that, say, the read thread is going to want to try and
// access the write store.
//---------------------------------------------------------------------------
#include <vcl.h>
#include <stdio.h>
#pragma hdrstop

#include "SelfImage_Utility.h"
#include "SelfImage_Exceptions.h"
#include "SelfImage_TImageStore.h"
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// TImageStore constructor that opens the target at the same time
//
__fastcall TImageStore::TImageStore(TImageStoreType isType, AnsiString Target, bool WriteAccess) {
  FType = isType;
  FFileName = Target;
  bWriteAccess = WriteAccess;
  if (!Open())
    throw ESelfImageError("Could not open store target \"" + Target + "\": " + GetLastErrorMessage());
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// TImageStore::~TImageStore()
//
__fastcall TImageStore::~TImageStore() {
  Close();
}  // __fastcall TImageStore::~TImageStore()
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Open an image store - this determines the type of the store and whether
// or not NT open functions are available and then decides the best way
// to open the target.
//
bool __fastcall TImageStore::Open(void) {

  switch (Type) {
    case isFile: {
      LARGE_INTEGER nnFileSize;

      DWORD Access     = GENERIC_READ | (bWriteAccess?GENERIC_WRITE:0);
      DWORD ShareMode  = (bWriteAccess?0:FILE_SHARE_READ);
      DWORD CreateMode = (bWriteAccess?CREATE_ALWAYS:OPEN_EXISTING);
      hHandle = CreateFile(FileName.c_str(), Access, ShareMode, NULL, CreateMode, FILE_ATTRIBUTE_NORMAL, NULL);
      if (hHandle != INVALID_HANDLE_VALUE) {
        if (!bWriteAccess) {
          nnFileSize.LowPart = GetFileSize(hHandle, (unsigned long *)&nnFileSize.HighPart);
          FGeometry.Bytes = nnFileSize.QuadPart;
          CheckCompression();
        } else
          FGeometry.Bytes = -1;
      }  // if (hHandle != INVALID_HANDLE_VALUE)
      break;
    }  // case isFile:
    case isDrive: {
      PARTITION_INFORMATION PartitionInfo;
      DISK_GEOMETRY DiskGeometry;
      DWORD Dummy;

      DWORD Access     = GENERIC_READ | (bWriteAccess?GENERIC_WRITE:0);
      DWORD ShareMode  = (bWriteAccess?0:FILE_SHARE_READ|FILE_SHARE_WRITE);
      if (HaveNTCalls)
        hHandle = NTOpen(FileName, SYNCHRONIZE | Access, FILE_ATTRIBUTE_NORMAL, ShareMode, FILE_OPEN, FILE_SYNCHRONOUS_IO_NONALERT | FILE_SEQUENTIAL_ONLY);
      else
        hHandle = CreateFile(FileName.c_str(), GENERIC_READ, ShareMode, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN | FILE_FLAG_BACKUP_SEMANTICS, NULL);
      if (hHandle != INVALID_HANDLE_VALUE) {
        DeviceIoControl(hHandle, IOCTL_DISK_GET_PARTITION_INFO, NULL, 0, &PartitionInfo, sizeof(PartitionInfo), &Dummy, NULL);
        DeviceIoControl(hHandle, IOCTL_DISK_GET_DRIVE_GEOMETRY	, NULL, 0, &DiskGeometry, sizeof(DiskGeometry), &Dummy, NULL);
        if (FileName.Pos("Floppy") || FileName.Pos("A:") || FileName.Pos("B:"))
          FGeometry.Bytes = DiskGeometry.Cylinders.QuadPart * DiskGeometry.TracksPerCylinder * DiskGeometry.SectorsPerTrack * DiskGeometry.BytesPerSector;
        else
          FGeometry.Bytes = PartitionInfo.PartitionLength.QuadPart;
        FGeometry.BytesPerSector = DiskGeometry.BytesPerSector;
        FGeometry.Sectors = Geometry.Bytes / Geometry.BytesPerSector;
      }  // if (hHandle != INVALID_HANDLE_VALUE)
      break;
    }  // case isDrive:
  }  // switch (Type)

  return (hHandle != INVALID_HANDLE_VALUE);
}  // bool __fastcall TImageStore::Open(void)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Close() - Close an image store
//
bool __fastcall TImageStore::Close(void) {
  return CloseHandle(hHandle);
}  // bool __fastcall Close(void)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Read() - As the name suggests, read from the file/device.  Special care
// has to be taken when reading from handles created with "NtCreateFile()".
// Windows versions based on NT have problems with partial reads of disk
// devices opened with NtCreateFile().  When there isn't enough data in the
// file to satisfy the entire length of the read, ReadFile() is supposed to
// always read as many as there are.  Disk devices of the type
// "\Device\Harddisk0\Partition1" opened with NtCreateFile don't do this.
// When there isn't enough data left to satisfy the entire read, ReadFile()
// will fail entirely and GetLastError() will return ERROR_INVALID_PARAMETER.
// To compensate for this, whenever we encounter ERROR_INVALID_PARAMETER, we
// step down the size of the read to the device's sector size and then read
// until we get an error again.
//
bool __fastcall TImageStore::Read(void *buffer, int nLength, unsigned *nBytesRead) {
  unsigned nBytesReadTemp, pos;
  bool bSuccess;
  bSuccess = ReadFile(hHandle, buffer, nLength, (unsigned long *)&nBytesReadTemp, NULL);
  if (!bSuccess && GetLastError() == ERROR_INVALID_PARAMETER) {
    // ERROR_INVALID_PARAMETER - step down read size as discussed above
    pos = 0;
    while (ReadFile(hHandle, (char *)buffer + pos, Geometry.BytesPerSector?Geometry.BytesPerSector:512, (unsigned long *)&nBytesReadTemp, NULL))
      pos += nBytesReadTemp;
    nBytesReadTemp = pos;
    bSuccess = true;
  }  // if (!bSuccess)
  *nBytesRead = nBytesReadTemp;
  return bSuccess;
}  // bool __fastcall TImageStore::Read(void *buffer, int nLength, unsigned *nBytesRead)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Write() - similar to Read() but different.
//
bool __fastcall TImageStore::Write(void *buffer, int nLength, unsigned *nBytesWritten) {
  return (WriteFile(hHandle, buffer, nLength, (unsigned long *)nBytesWritten, NULL));
}  // bool __fastcall TImageStore::Write(void *buffer, int nLength, unsigned *nBytesWritten)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
//  Set the file position pointer
//
bool __fastcall TImageStore::Seek(unsigned __int64 Offset, DWORD MoveMethod)
{
  LOW_DWORD(Offset) = SetFilePointer(hHandle, LOW_DWORD(Offset), &HIGH_DWORD(Offset), MoveMethod);
  if (Offset == 0x00000000FFFFFFFFui64)
    return (GetLastError() == NO_ERROR);
  else
    return true;
}  // bool __fastcall TImageStore::Seek(__int64)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Check to see if the file is compressed.  This is a full check that looks
// at the contents of the file to see if it contains the correct signature
// and also looks to see if it's in SelfImage's block format for parallel
// decompression.
//
void __fastcall TImageStore::CheckCompression(void)
{
  unsigned char *Buffer = NULL;
  AnsiString Extension = ExtractFileExt(FileName);

  try {
    if (Extension == ".gz") {
      unsigned ReadCount;
      Buffer = new unsigned char[20];
      // Read in the gzip header - 20 bytes is all we need
      Read(Buffer, 20, &ReadCount);
      if (ReadCount == 20) {
        // If the file is actually a gzip file then the first two bytes will be 0x1F 0x8B.  The next byte identifies the
        // gzip compression method - since deflate is all we support, we check to make sure it's a deflate gzip file.
        // Deflate is all zlib supports right now - probably all it will ever support.
        if (Buffer[0] == 0x1F && Buffer[1] == 0x8B && Buffer[2] == 0x08)
          // By this point we know for sure it's a gzip file.  Now check to see if it's a gzip that's in SelfImage block format.
          // If it is, it will have a gzip Extra field identified by 'SI' followed by the field length, which must be little-endian 0x0008
          if (Buffer[12] == 'S' && Buffer[13] == 'I' && Buffer[14] == 0x08 && Buffer[15] == 0x00)
            FCompressionFormat = cfGzipBlock;
          else
            FCompressionFormat = cfGzipStream;
      }  // if (ReadCount == 20)
    }  // if (Extension == ".gz")
    else if (Extension == ".bz2") {
      unsigned ReadCount;
      Buffer = new unsigned char[1024 * 1024];
      // Need to read in the entire first bzip2 block, plus the header of the next one.
      // Techinically that's 909610 bytes - make it one meg for good measure
      if (Read(Buffer, 1024 * 1024, &ReadCount)) {
        // If it's a bz2 file, it will start with "BZh" and then have a number from 1 to 9
        if (ReadCount > 20 && Buffer[0] == 'B' && Buffer[1] == 'Z' && Buffer[2] == 'h' && Buffer[3] >= '1' && Buffer[3] <= '9') {
           char *Header = "BZh_1AY&SY";
           Header[3] = Buffer[3];
           // Check to see if there is another full bzip2 header after the first bzip2 block.  If there is,
           // then we have a SelfImage format file - otherwise it's a standard bzip2 file with blocks that
           // we can't tell the size of.
           if (FindString(Buffer+10, ReadCount, Header, true))
             FCompressionFormat = cfBzip2Block;
           else
             FCompressionFormat = cfBzip2Stream;
        }  // if (...is a bzip2 file)
      }  // if (Read(Buffer, 20, &ReadCount))
      // Read in the first compression block
    }  // else if (Extension == ".bz2")
  }  // try
  __finally {
    if (Buffer)
      delete [] Buffer;
    Seek(0, FILE_BEGIN);
  }  // __finally
}  // void __fastcall TImageStore::CheckCompression(void)
//---------------------------------------------------------------------------


